from threading import Thread
import re

words = {}
current_jobs = {}
matcher = re.compile(r'[\w-]{3,}')

def process(filename, data):
    result = {}
    
    for w in matcher.findall(data):
        try:
            result[w] += 1
        except KeyError:
            result[w] = 1

    for w, c in result.iteritems():
        words.setdefault(w, {})[filename] = c
    
    for w in [w for w, files in words.iteritems() if w not in result and filename in files]:
        del words[w][filename]

def add_job(filename, data):
    if filename in current_jobs:
        current_jobs[filename].join()

    thread = Thread(target=process, args=(filename, data))
    current_jobs[filename] = thread
    thread.start()
    


import anydbm
import os.path
import os
import json

default_prefs = {
    'default': {
        'font': 'Monospace 11',
        'use-tabs': True,
        'tab-width': 4,
        'show-right-margin': False,
        'right-margin': 100,
        'show-line-numbers': True,
        'wrap-text': False,
        'style': 'classic',
        'auto-indent': True,
        'indent-on-tab': True,
        'smart-home-end': True,
        'highlight-current-line': True,
        'show-whitespace': False,
    },
    'python': {
        'use-tabs': False,
        'show-right-margin': True,
    }
}

registered_dialogs = {}

def register_dialog(name, callback, *keywords):
    registered_dialogs[name] = keywords, callback

def load_json_settings(name, default=None):
    filename = get_settings_path(name)
    try:
        with open(filename) as f:
            try:
                return json.load(f)
            except ValueError:
                pass
    except IOError:
        pass

    return default

def save_json_settings(name, value):
    filename = get_settings_path(name)
    with open(filename, 'w') as f:
        json.dump(value, f, sort_keys=True, indent=4)


class CompositePreferences(object):
    def __init__(self, *prefs):
        self.prefs = prefs
    
    def __getitem__(self, key):
        for p in self.prefs:
            try:
                return p[key]
            except KeyError:
                pass
                
        raise KeyError('There is no %s in preferences' % key)

def get_settings_path(name):
    path = os.path.join(os.path.expanduser("~"), '.local', 'snaked')
    if not os.path.exists(path):
        os.makedirs(path, mode=0755)
        
    return os.path.join(path, name)
    

class KVSettings(object):
    def __init__(self, name):
        self.db = anydbm.open(get_settings_path(name), 'c')
    
    def __getitem__(self, key):
        return self.db[key]
    
    def __contains__(self, key):
        return self.db.has_key(key)
    
    def __setitem__(self, key, value):
        self.db[key] = value
    
    def __del__(self):
        self.db.close()

class ListSettings(object):
    def __init__(self, name):
        self.path = get_settings_path(name)

    def exists(self):
        return os.path.exists(self.path)
    
    def load(self):
        try:
            return [l.strip() for l in open(self.path)]
        except IOError:
            return []
    
    def store(self, data):
        open(self.path, 'w').write('\n'.join(data))



import os.path
import weakref

from snaked.util import idle, join_to_file_dir, BuilderAware, open_mime, refresh_gui, single_ref
from snaked.core.shortcuts import ShortcutActivator
from snaked.core.prefs import ListSettings

import settings
import searcher

class QuickOpenDialog(BuilderAware):
    def __init__(self):
        super(QuickOpenDialog, self).__init__(join_to_file_dir(__file__, 'gui.glade'))
        self.shortcuts = ShortcutActivator(self.window)
        self.shortcuts.bind('Escape', self.hide)
        self.shortcuts.bind('<alt>Up', self.project_up)
        self.shortcuts.bind('<alt>Down', self.project_down)
        self.shortcuts.bind('Return', self.open_file)
        self.shortcuts.bind('<ctrl>Return', self.open_mime)
        self.shortcuts.bind('<alt>s', self.focus_search)

    @single_ref
    def prefs(self):
        return ListSettings('project-roots.db')
        
    def show(self, editor):
        self.editor = weakref.ref(editor)
        
        self.update_recent_projects()
        self.update_projects(editor.project_root)

        self.search_entry.grab_focus()
        
        editor.request_transient_for.emit(self.window)
        self.window.present()
    
    def update_recent_projects(self):
        saved_projects = self.prefs.load()
                
        if any(p not in saved_projects for p in settings.recent_projects):
            [saved_projects.append(p) for p in settings.recent_projects
                if p not in saved_projects]
            self.prefs.store(saved_projects)
            settings.recent_projects = saved_projects
            return
            
        if any(p not in settings.recent_projects for p in saved_projects):
            [settings.recent_projects.append(p) for p in saved_projects
                if p not in settings.recent_projects]
    
    def update_projects(self, root):
        self.projects_cbox.set_model(None)
        self.projectlist.clear()
        
        index = 0
        for i, r in enumerate(settings.recent_projects):
            if r == root:
                index = i
            self.projectlist.append((r,))
        
        self.projects_cbox.set_model(self.projectlist)
        self.projects_cbox.set_active(index)
    
    def hide(self):
        self.window.hide()
        
    def on_delete_event(self, *args):
        self.hide()
        return True
    
    def project_up(self):
        idx = self.projects_cbox.get_active()
        idx = ( idx - 1 ) % len(self.projectlist)
        self.projects_cbox.set_active(idx)
        
    def project_down(self):
        idx = self.projects_cbox.get_active()
        idx = ( idx + 1 ) % len(self.projectlist)
        self.projects_cbox.set_active(idx)

    def get_current_root(self):
        return self.projectlist[self.projects_cbox.get_active()][0]
    
    def fill_filelist(self, search):
        self.filelist.clear()
        
        current_search = object()
        self.current_search = current_search
        
        already_matched = {}
        i = 0
        
        for m in (searcher.name_start_match, searcher.name_match,
                searcher.path_match, searcher.fuzzy_match):
            for p in searcher.search(self.get_current_root(), '', m(search), already_matched):
                if self.current_search is not current_search:
                    return
                    
                already_matched[p] = True            
                self.filelist.append(p)
                
                if i % 10 == 0:
                    refresh_gui()
                    
                i += 1

        self.filelist_tree.columns_autosize()

    def on_search_entry_changed(self, *args):
        search = self.search_entry.get_text().strip()
        if search:
            idle(self.fill_filelist, search)
        
    def on_projects_cbox_changed(self, *args):
        self.on_search_entry_changed()

    def get_selected_file(self):
        (model, iter) = self.filelist_tree.get_selection().get_selected()
        if iter:
            return os.path.join(self.get_current_root(), *self.filelist.get(iter, 1, 0))
        else:
            return None
    
    def open_file(self):
        fname = self.get_selected_file()
        if fname:
            self.hide()
            refresh_gui()
            self.editor().open_file(fname)
        
    def open_mime(self):
        fname = self.get_selected_file()
        if fname:
            self.hide()
            refresh_gui()
            open_mime(fname)

    def focus_search(self):
        self.search_entry.grab_focus()


import weakref

import gtksourceview2

from snaked.util import BuilderAware, join_to_file_dir, idle
from snaked.core import shortcuts
import snaked.core.prefs as prefs


class PreferencesDialog(BuilderAware):
    def __init__(self, prefs):
        BuilderAware.__init__(self, join_to_file_dir(__file__, 'editor_prefs.glade'))
        self.activator = shortcuts.ShortcutActivator(self.window)
        self.activator.bind('Escape', self.hide)

        self.prefs = prefs
        self.original_prefs = prefs.copy()

        self.langs.append(('default', ))
        lm = gtksourceview2.language_manager_get_default()
        for lang_id in sorted(lm.get_language_ids()):
            self.langs.append((lang_id, ))

        sm = gtksourceview2.style_scheme_manager_get_default()
        for style_id in sm.get_scheme_ids():
            self.styles.append((style_id, ))
            
        self.checks = ['use-tabs', 'show-right-margin', 'show-line-numbers', 'wrap-text',
            'highlight-current-line', 'show-whitespace']

        for name in self.checks:
            getattr(self, name.replace('-', '_')).connect(
                'toggled', self.on_checkbox_toggled, name)

        self.margin_width.connect('value-changed', self.on_spin_changed, 'right-margin')
        self.tab_width.connect('value-changed', self.on_spin_changed, 'tab-width')

        self.font.connect('font-set', self.on_font_set, 'font')
        
    def show(self, editor):
        self.editor = weakref.ref(editor)        
        editor.request_transient_for.emit(self.window)
        self.select_lang(editor.lang)
        self.refresh_lang_settings()
        self.window.present()

    def select_lang(self, lang_id):
        for i, (name,) in enumerate(self.langs):
            if name == lang_id:
                self.langs_view.set_cursor((i,))
                self.langs_view.scroll_to_cell((i,), None, True, 0.5, 0)
                return 
                
        self.langs_view.set_cursor((0,))
    
    def get_current_lang_id(self):
        (model, iter) = self.langs_view.get_selection().get_selected()
        return self.langs.get_value(iter, 0)
    
    def get_lang_prefs(self, lang_id):
        if lang_id == 'default':
            return prefs.CompositePreferences(self.prefs.get('default', {}),
                prefs.default_prefs.get(lang_id, {}), prefs.default_prefs['default'])
        else:
            return prefs.CompositePreferences(self.prefs.get(lang_id, {}),
                self.prefs.get('default', {}), prefs.default_prefs.get(lang_id, {}),
                prefs.default_prefs['default'])

    def get_default_lang_prefs(self, lang_id):
        if lang_id == 'default':
            return prefs.CompositePreferences(prefs.default_prefs.get(lang_id, {}),
                prefs.default_prefs['default'])
        else:
            return prefs.CompositePreferences(self.prefs.get('default', {}),
                prefs.default_prefs.get(lang_id, {}), prefs.default_prefs['default'])
     
    def set_checkbox(self, pref, name):
        getattr(self, name.replace('-', '_')).set_active(pref[name])
        
    def refresh_lang_settings(self, *args):
        pref = self.get_lang_prefs(self.get_current_lang_id())
        
        for name in self.checks:
            self.set_checkbox(pref, name)        
        
        self.select_style(pref['style'])
        
        self.margin_width.set_value(pref['right-margin'])
        self.tab_width.set_value(pref['tab-width'])
            
        self.font.set_font_name(pref['font'])
        
    def select_style(self, style_id, try_classic=True):
        for i, (name,) in enumerate(self.styles):
            if name == style_id:
                self.style_cb.set_active(i)
                return 
        
        if try_classic:
            return self.select_style('classic', False)

        self.style_cb.set_active(0)

    def hide(self):
        self.save_settings()
        self.editor().message('Editor settings saved')
        self.window.destroy()

    def save_settings(self):
        prefs.save_json_settings('langs.conf', self.prefs)

    def on_delete_event(self, *args):
        idle(self.hide)
        return True

    def update_pref_value(self, lang_id, name, value):
        current_value = self.get_lang_prefs(lang_id)[name]
        if value != current_value:
            default_value = self.get_default_lang_prefs(lang_id)[name]
            if value == default_value:
                try:
                    del self.prefs[lang_id][name]
                except KeyError:
                    pass
            else:
                self.prefs.setdefault(lang_id, {})[name] = value
            
            self.editor().settings_changed.emit()

    def on_style_cb_changed(self, *args):
        (style_id,) = self.styles[self.style_cb.get_active()]
        lang_id = self.get_current_lang_id()
        self.update_pref_value(lang_id, 'style', style_id)
        
    def on_checkbox_toggled(self, widget, name):
        self.update_pref_value(self.get_current_lang_id(), name, widget.get_active())

    def on_spin_changed(self, widget, name):
        self.update_pref_value(self.get_current_lang_id(), name, widget.get_value_as_int())

    def on_font_set(self, widget, name):
        self.update_pref_value(self.get_current_lang_id(), name, widget.get_font_name())

    def on_reset_to_default_clicked(self, *args):
        try:
            del self.prefs[self.get_current_lang_id()]
        except KeyError:
            pass
        
        self.refresh_lang_settings()
        self.editor().settings_changed.emit()



import weakref

from snaked.util import BuilderAware, join_to_file_dir, idle

import snaked.core.prefs as prefs

class PreferencesDialog(BuilderAware):
    def __init__(self):
        BuilderAware.__init__(self, join_to_file_dir(__file__, 'prefs.glade'))
        
        from snaked.core.shortcuts import ShortcutActivator
        self.activator = ShortcutActivator(self.window)
        self.activator.bind('Escape', self.hide)
        self.activator.bind('Return', self.activate)
        self.activator.bind('<alt>s', self.focus_search)
        
    def hide(self):
        self.window.destroy()
        
    def show(self, editor):
        self.editor = weakref.ref(editor)
        self.fill_dialogs(None)
        editor.request_transient_for.emit(self.window)
        self.window.show()
        
    def fill_dialogs(self, search):
        self.dialogs.clear()

        for name, (keywords, show_func) in prefs.registered_dialogs.iteritems():
            if not search or any(w.startswith(search) for w in keywords):
                markup = '<b>%s</b>\n<small>%s</small>' % (
                    name, u' \u2022 '.join(keywords))
                self.dialogs.append((name, markup))
        
    def on_delete_event(self, *args):
        return False
        
    def on_search_entry_changed(self, *args):
        search = self.search_entry.get_text().strip().lower()
        idle(self.fill_dialogs, search)
        
    def activate(self):
        (model, iter) = self.dialogs_view.get_selection().get_selected()
        name = model.get_value(iter, 0)
        prefs.registered_dialogs[name][1](self.editor())
        idle(self.hide)
        
    def focus_search(self):
        self.search_entry.grab_focus()


import gtk
import weakref
from gobject import timeout_add, source_remove

class FeedbackPopup(object):
    def __init__(self, view):
        self.view = weakref.ref(view)

        self.bar = gtk.EventBox()

        box = gtk.EventBox()
        box.set_border_width(5)
        
        self.label = gtk.Label()
        box.add(self.label)
        box.show_all()
        
        self.bar.add(box)
        
        view.add_child_in_window(self.bar, gtk.TEXT_WINDOW_TEXT, 0, 0)
        
        self.timeout_id = None
    
    def remove_timeout(self):
        if self.timeout_id:
            source_remove(self.timeout_id)
        
        self.timeout_id = None        
    
    def show(self, text, timeout=1500):
        self.remove_timeout()
        
        self.label.set_text(u'  '+ text + u'  ')
        
        x, y, w, h, d = self.view().get_window(gtk.TEXT_WINDOW_TEXT).get_geometry()
        sw, sh = self.bar.size_request()
        x, y = w - sw, h - sh        
        self.view().move_child(self.bar, x, y) 
        self.bar.show()
        
        self.timeout_id = timeout_add(timeout, self.hide)

    def hide(self):
        self.remove_timeout()
        self.bar.hide()
        return False



import os.path
import weakref

import gtk
import gtksourceview2
import pango

from ..util import save_file, idle, get_project_root, lazy_property
from ..signals import SignalManager, Signal, connect_all, connect_external

import prefs
from .shortcuts import register_shortcut, load_shortcuts
from .plugins import PluginManager

class Editor(SignalManager):
    editor_closed = Signal()
    request_to_open_file = Signal(str, return_type=object)
    request_close = Signal()
    settings_changed = Signal()
    get_title = Signal(return_type=str)
    get_file_position = Signal(return_type=int)
    before_close = Signal()
    file_loaded = Signal()
    change_title = Signal(str) 
    request_transient_for = Signal(object)
    file_saved = Signal()
    push_escape_callback = Signal(object, object)
    plugins_changed = Signal()

    def __init__(self):    
        self.uri = None
        
        sw = gtk.ScrolledWindow()
        sw.set_policy(gtk.POLICY_AUTOMATIC, gtk.POLICY_ALWAYS)
                
        self.buffer = gtksourceview2.Buffer()

        self.view = gtksourceview2.View()
        self.view.set_buffer(self.buffer)
        sw.add(self.view)


        self.widget = gtk.VBox(False, 0)
        self.widget.pack_start(sw)

        self.widget.show_all()
        
        connect_all(self, buffer=self.buffer)

    def update_title(self):
        modified = '*' if self.buffer.get_modified() else ''
        
        if self.uri:
            title = self.get_title.emit()

            if not title:
                title = os.path.basename(self.uri)
        else:
            title = 'Unknown'
        
        self.change_title.emit(modified + title)          

    def load_file(self, filename):
        self.uri = os.path.abspath(filename)
        
        if os.path.exists(self.uri):
            self.view.window.freeze_updates()        
            self.on_modified_changed_handler.block()
            self.buffer.begin_not_undoable_action()
            
            self.buffer.set_text(open(filename).read().decode('utf-8'))
            self.buffer.set_modified(False)
            
            self.buffer.end_not_undoable_action()
            self.on_modified_changed_handler.unblock()
            self.view.window.thaw_updates()
            
            pos = self.get_file_position.emit()
            if pos is not None and pos >= 0:
                self.buffer.place_cursor(self.buffer.get_iter_at_line(pos))
                self.view.scroll_to_mark(self.buffer.get_insert(), 0.001, use_align=True, xalign=1.0)
            else:
                self.buffer.place_cursor(self.buffer.get_start_iter())
                
        self.file_loaded.emit()
                
        self.update_title()
        
    @connect_external('buffer', 'modified-changed', idle=True)
    def on_modified_changed(self, *args):
        self.update_title()
            
    def save(self):
        if self.uri:
            try:
                save_file(self.uri, self.buffer.get_text(*self.buffer.get_bounds()), 'utf-8')
                if not self.buffer.get_modified():
                    self.message("%s saved" % self.uri)
                self.buffer.set_modified(False)
                self.file_saved.emit()
            except Exception, e:
                self.message(str(e), 5000)

    @property
    def project_root(self):
        if self.uri:
            root = get_project_root(self.uri)
            if not root:
                root = os.path.dirname(self.uri)
          
            return root 
            
        return None

    def open_file(self, filename):        
        editor = self.request_to_open_file.emit(filename)
        if not self.uri and not self.buffer.get_modified() and self.text == u'':
            self.request_close.emit()
            
        return editor

    @property
    def cursor(self):
        return self.buffer.get_iter_at_mark(self.buffer.get_insert())

    @property
    def text(self):
        return self.buffer.get_text(*self.buffer.get_bounds())

    def goto_line(self, line):
        iterator = self.buffer.get_iter_at_line(line - 1)
        self.buffer.place_cursor(iterator)
        self.view.scroll_to_iter(iterator, 0.001, use_align=True, xalign=1.0)

    @lazy_property
    def feedback_popup(self):
        from .feedback import FeedbackPopup
        return FeedbackPopup(self.view)
                
    def message(self, message, timeout=1500):
        popup = self.feedback_popup
        popup.show(message, timeout)
        self.push_escape(popup.hide)

    def push_escape(self, callback, *args):
        self.push_escape_callback.emit(callback, args)
        
class EditorManager(object):
    def __init__(self):
        self.editors = []
        self.style_manager = gtksourceview2.style_scheme_manager_get_default()
        self.lang_manager = gtksourceview2.language_manager_get_default()
        
        self.plugin_manager = PluginManager()
        prefs.register_dialog('Plugins', self.plugin_manager.show_plugins_prefs, 'plugin',
            'extension')
            
        prefs.register_dialog('Key configuration', self.show_key_preferences, 'key', 'bind', 'shortcut')

        prefs.register_dialog('Editor settings', self.show_editor_preferences,
            'editor', 'font', 'style', 'margin', 'line', 'tab', 'whitespace')
        
        self.escape_stack = []
        self.escape_map = {}
        
        self.session = None

        load_shortcuts()
        self.register_app_shortcuts()

    def register_app_shortcuts(self):
        register_shortcut('quit', '<ctrl>q', 'Application', 'Quit')        
        register_shortcut('close-window', '<ctrl>w', 'Window', 'Closes window')
        register_shortcut('save', '<ctrl>s', 'File', 'Saves file')
        register_shortcut('next-editor', '<alt>Right', 'Window', 'Switches to next editor')
        register_shortcut('prev-editor', '<alt>Left', 'Window', 'Switches to previous editor')
        register_shortcut('new-file', '<ctrl>n', 'File',
            'Open dialog to choose new file directory and name')
        register_shortcut('show-preferences', '<ctrl>p', 'Window', 'Open preferences dialog')
        
    def open(self, filename):
        editor = Editor()
        self.editors.append(editor)
        
        connect_all(self, editor)

        idle(self.set_editor_prefs, editor, filename)
        idle(self.set_editor_shortcuts, editor)
        idle(self.plugin_manager.editor_created, editor)
        
        self.manage_editor(editor)
        
        if filename:
            idle(editor.load_file, filename)

        idle(self.plugin_manager.editor_opened, editor)
        
        return editor

    @lazy_property
    def lang_prefs(self):
        return prefs.load_json_settings('langs.conf', {})
    
    def set_editor_prefs(self, editor, filename):
        if filename:
            lang = self.lang_manager.guess_language(filename, None)
            editor.buffer.set_language(lang)
        else:
            lang = None
            
        editor.lang = lang.get_id() if lang else 'default'
        
        pref = prefs.CompositePreferences(self.lang_prefs.get(editor.lang, {}),
            self.lang_prefs.get('default', {}), prefs.default_prefs.get(editor.lang, {}),
            prefs.default_prefs['default'])
        
        style_scheme = self.style_manager.get_scheme(pref['style'])
        editor.buffer.set_style_scheme(style_scheme)
        
        font = pango.FontDescription(pref['font'])
        editor.view.modify_font(font)
        
        editor.view.set_auto_indent(pref['auto-indent'])
        editor.view.set_indent_on_tab(pref['indent-on-tab'])
        editor.view.set_insert_spaces_instead_of_tabs(not pref['use-tabs'])
        editor.view.set_smart_home_end(pref['smart-home-end'])
        editor.view.set_highlight_current_line(pref['highlight-current-line'])
        editor.view.set_show_line_numbers(pref['show-line-numbers'])
        editor.view.set_tab_width(pref['tab-width'])
        editor.view.set_draw_spaces(pref['show-whitespace'])
        editor.view.set_right_margin_position(pref['right-margin'])
        editor.view.set_show_right_margin(pref['show-right-margin'])
        editor.view.set_wrap_mode(gtk.WRAP_WORD if pref['wrap-text'] else gtk.WRAP_NONE)

    @Editor.editor_closed(idle=True)
    def on_editor_closed(self, editor):
        if self.session and len(self.editors) == 1:
            self.save_session(self.session)
            self.session = None
        
        self.plugin_manager.editor_closed(editor)
        self.editors.remove(editor)
        
        if not self.editors:
            self.plugin_manager.quit()
            gtk.main_quit()

    @Editor.change_title
    def on_editor_change_title(self, editor, title):
        self.set_editor_title(editor, title)

    @Editor.request_close
    def on_editor_close_request(self, editor):
        self.close_editor(editor)
    
    @Editor.request_to_open_file
    def on_request_to_open_file(self, editor, filename):
        for e in self.editors:
            if e.uri == filename:
                self.focus_editor(e)
                break
        else:
            e = self.open(filename)
        
        return e

    @Editor.request_transient_for
    def on_request_transient_for(self, editor, window):
        self.set_transient_for(editor, window)

    @Editor.settings_changed(idle=True)
    def on_editor_settings_changed(self, editor):
        self.set_editor_prefs(editor, editor.uri)
        for e in self.editors:
            if e is not editor:
                idle(self.set_editor_prefs, e, e.uri)
        
    def new_file_action(self, editor):
        from snaked.core.gui import new_file
        new_file.show_create_file(editor)
        
    def quit(self, *args):
        if self.session:
            self.save_session(self.session)
            self.session = None
            
        [self.close_editor(e) for e in self.editors]

    def get_session_files(self, session):
        self.session = session
        from .prefs import ListSettings
        settings = ListSettings('session-%s.db' % session)
        return settings.load()
                    
    def save_session(self, session):
        from .prefs import ListSettings
        settings = ListSettings('session-%s.db' % session)
        settings.store(e.uri for e in self.editors if e.uri)

    @Editor.push_escape_callback
    def on_push_escape_callback(self, editor, callback, args):
        key = (callback,) + tuple(map(id, args))
        if key in self.escape_map:
            return
            
        self.escape_map[key] = True
        self.escape_stack.append((key, callback, map(weakref.ref, args)))
        
    def process_escape(self, editor):
        while self.escape_stack:
            key, cb, args = self.escape_stack.pop()
            del self.escape_map[key]
            realargs = [a() for a in args]
            if not any(a is None for a in realargs):
                cb(*realargs)
                return False
                
        return False

    def show_key_preferences(self, editor):
        from snaked.core.gui.shortcuts import ShortcutsDialog
        dialog = ShortcutsDialog()
        dialog.show(editor)
        
    def show_preferences(self, editor):
        from snaked.core.gui.prefs import PreferencesDialog
        dialog = PreferencesDialog()
        dialog.show(editor)
        
    def show_editor_preferences(self, editor):
        from snaked.core.gui.editor_prefs import PreferencesDialog
        dialog = PreferencesDialog(self.lang_prefs)
        dialog.show(editor)

    @Editor.plugins_changed
    def on_plugins_changed(self, editor):
        for e in self.editors:
            self.set_editor_shortcuts(e)


from util import (iter_lines, line_is_empty, get_next_not_empty_line, get_whitespace,
    line_text, next_line, prev_line, cursor_on_start_or_end_whitespace, get_line_bounds)

def extend_with_gap(from_iter, ws, delta):
    n = None
    for p, n in iter_lines(from_iter, delta):
        if line_is_empty(n):
            ne = get_next_not_empty_line(n, delta)
            if ne and len(get_whitespace(ne)) >= ws:
                n.set_line(ne.get_line())
            else:
                return p
    
        n_ws = len(get_whitespace(n))         
        if n_ws < ws:
            return p
    
    return n if n else from_iter.copy()

def extend_without_gap(from_iter, ws, delta):
    n = None
    for p, n in iter_lines(from_iter, delta):
        if line_is_empty(n):
            ne = get_next_not_empty_line(n, delta)
            if ne and delta > 0 and len(get_whitespace(ne)) > ws:
                n.set_line(ne.get_line())
            else:
                return p
    
        n_ws = len(get_whitespace(n))         
        if n_ws < ws:
            return p
    
    return n if n else from_iter.copy()

def extend_block_without_gap(from_iter, ws, delta):
    n = None
    for p, n in iter_lines(from_iter, delta):
        if line_is_empty(n):
            ne = get_next_not_empty_line(n, delta)
            if ne:
                n.set_line(ne.get_line())
            else:
                return p
    
        n_ws = len(get_whitespace(n))         
        
        if n_ws < ws or ( n_ws == ws and len(line_text(n).strip()) > 4 ):
            return p
    
    return n if n else from_iter.copy()

def block_smart_extend(has_selection, start, end):
    end = end.copy()
    if not end.is_end():
        end.backward_lines(1)
    
    start_ws = len(get_whitespace(start))
    prev_empty = start.is_start() or line_is_empty(prev_line(start))
    prev_ws = len(get_whitespace(prev_line(start)))
    
    end_ws = len(get_whitespace(end))
    next_empty = end.is_end() or line_is_empty(next_line(end))
    next_ws = len(get_whitespace(next_line(end)))

    newstart, newend = start.copy(), end
    
    if not has_selection and start.get_line() == end.get_line() and \
            ( next_empty or next_ws < end_ws ) and (prev_empty or prev_ws < start_ws):
        pass   
    elif not prev_empty and not next_empty and prev_ws == start_ws == next_ws == end_ws:
        newstart = extend_without_gap(start, start_ws, -1)
        newend = extend_without_gap(end, end_ws, 1)
    elif prev_empty and not next_empty and next_ws == end_ws:
        newend = extend_without_gap(end, end_ws, 1)
    elif not prev_empty and next_empty and prev_ws == start_ws:
        newstart = extend_without_gap(start, start_ws, -1)
    elif not next_empty and next_ws > start_ws:
        newend = extend_block_without_gap(end, start_ws, 1)
    elif ( not next_empty and next_ws == start_ws ) or ( not prev_empty and prev_ws >= start_ws ):
        if not prev_empty:
            newstart = extend_without_gap(start, start_ws, -1)
        if not next_empty:
            newend = extend_without_gap(end, start_ws, 1)
    elif next_empty and prev_empty:
        newstart = extend_with_gap(start, start_ws, -1)
        newend = extend_with_gap(end, start_ws, 1)
    elif next_empty and not prev_empty and prev_ws < start_ws:
        newend = extend_with_gap(end, start_ws, 1)
    
    if has_selection and start.equal(newstart) and end.equal(newend):
        if not prev_empty:
            newstart.backward_lines(1)
        else:
            ne = get_next_not_empty_line(start, -1)
            if ne:
                newstart = ne
        
        if not next_empty and len(line_text(next_line(end)).strip()) < 5:
            newend.forward_lines(1)        
            
    newend.forward_lines(1)
    return newstart, newend
    
def get_smart_select(editor):
    if editor.buffer.get_has_selection():
        start, end = editor.buffer.get_selection_bounds()
        if start.starts_line() and ( end.starts_line() or end.is_end() ):
            return block_smart_extend(True, start, end)
        else:
            return line_smart_extend(True, start, end)
    else:
        cursor = editor.cursor
    
        if cursor_on_start_or_end_whitespace(cursor):
            return block_smart_extend(False, *get_line_bounds(cursor))
        else:
            return line_smart_extend(False, cursor, cursor.copy())     

def get_words_bounds(cursor, include_hyphen=False):
    return backward_word_start(cursor, include_hyphen), forward_word_end(cursor, include_hyphen)

def char_is_word(char, include_hyphen=False):
    return char and char.isalnum() or char == u'_' or (include_hyphen and char == u'-')

def backward_word_start(iter, include_hyphen=False):
    iter = iter.copy()
    iter.backward_char()
    while char_is_word(iter.get_char(), include_hyphen):
        iter.backward_char()
    
    iter.forward_char()
    
    return iter
    
def forward_word_end(iter, include_hyphen=False):
    iter = iter.copy()
    while char_is_word(iter.get_char()):
        iter.forward_char()
    
    return iter
    
def line_smart_extend(has_selection, start, end):
    import pairs_parser

    def ahtung():
        start.set_line(start.get_line())
        if not end.starts_line():
            end.set_line(end.get_line() + 1)
        
        return start, end

    left = start.copy()
    left.backward_chars(3)
    lchars = left.get_text(start)
    
    right = end.copy()
    right.forward_chars(3)
    rchars = end.get_text(right)

    if not rchars:
        rchars = [None]
        
    if not lchars:
        lchars = [None]

    text = start.get_buffer().get_text(*start.get_buffer().get_bounds())
    br, spos, epos = pairs_parser.get_brackets(text, start.get_offset())
    in_quotes = br in ('"', "'", '"""', "'''")
    #print br, spos, epos, start.get_offset()
    
    if not in_quotes and rchars[0] in (u'(', u'[', "'", '"'):
        try:
            br, spos, epos = pairs_parser.get_brackets(text, end.get_offset() + 1)
        except TypeError:
            br = None
            
        if not br: return ahtung()
        
        end.set_offset(epos)    
    elif char_is_word(lchars[-1]) or char_is_word(rchars[0]):
        start = backward_word_start(start, in_quotes)
        end = forward_word_end(end, in_quotes)
    elif lchars[-1] == u'.' or rchars[0] == u'.':
        if lchars[-1] == u'.':
            start.backward_char()
            start = backward_word_start(start)
        if rchars[0] == u'.':
            end.forward_char()
            end = forward_word_end(end)
    else:
        if not br: return ahtung()

        ostart = start.copy()
        oend = end.copy()
        
        start.set_offset(spos)
        end.set_offset(epos - 1)
        
        if ostart.equal(start) and oend.equal(end):
            start.backward_chars(len(br))
            end.forward_chars(len(br))
        
    return start, end      



author = 'Anton Bobrov<bobrov@vl.ru>'
name = 'Python support'
desc = 'Autocompletion, definitions navigation and smart ident'

langs = ['python']

handlers = {}
outline_dialog = None

def init(manager):
    manager.add_shortcut('python-goto-definition', 'F3', 'Python',
        'Navigates to python definition', goto_definition)

    manager.add_shortcut('python-outline', '<ctrl>o', 'Python',
        'Opens outline dialog', open_outline)

def editor_created(editor):
    editor.connect('get-title', on_editor_get_title)
    
def editor_opened(editor):
    from plugin import Plugin    
    h = Plugin(editor)
    handlers[editor] = h

def editor_closed(editor):
    handlers[editor].close()
    del handlers[editor]

def quit():
    global outline_dialog
    if outline_dialog:
        outline_dialog.window.destroy()
        del outline_dialog

def goto_definition(editor):
    try:
        h = handlers[editor]
    except KeyError:
        return
        
    h.goto_definition()

def on_editor_get_title(editor):
    if editor.uri.endswith('.py'):
        return get_python_title(editor.uri)
    
    return None

def get_python_title(uri):
    from os.path import dirname, basename, exists, join
    
    title = basename(uri)
    packages = []
    while True:
        path = dirname(uri)
        if path == uri:
            break
        
        uri = path
        
        if exists(join(uri, '__init__.py')):
            packages.append(basename(uri))
        else:
            break
            
    if packages:
        if title != '__init__.py':
            packages.insert(0, title.partition('.py')[0])
            
        return '.'.join(reversed(packages))
    else:
        return None

def open_outline(editor):
    global outline_dialog
    if not outline_dialog:
        from outline import OutlineDialog
        outline_dialog = OutlineDialog()
    
    outline_dialog.show(editor)


import re

from gtksourceview2 import CompletionProvider, CompletionProposal
from gtksourceview2 import COMPLETION_ACTIVATION_USER_REQUESTED

import gobject
from glib import markup_escape_text

pydoc_converter = re.compile(r'([^\n])\n[ \t]*?([^\n])')


class Proposal(gobject.GObject, CompletionProposal):
    def __init__(self, proposal):
        gobject.GObject.__init__(self)
        self.proposal = proposal
        
    def do_get_label(self):
        return self.proposal.name

    def do_get_text(self):
        return self.proposal.name

    def do_get_info(self):
        info = self.proposal.get_doc()
        if info:
            return markup_escape_text(pydoc_converter.sub(r'\1 \2', info))
        else:
            return ''                
                
class RopeCompletionProvider(gobject.GObject, CompletionProvider):
    def __init__(self, plugin):
        gobject.GObject.__init__(self)
        self.plugin = plugin
    
    def do_get_name(self):
        return 'python'
                
    def do_get_priority(self):
        return 1

    def do_set_priority(self):
        pass

    def do_get_activation(self):
        return COMPLETION_ACTIVATION_USER_REQUESTED 

    def do_populate(self, context):
        project = self.plugin.project
    
        from rope.contrib import codeassist
        
        try:
            proposals = codeassist.sorted_proposals(
                codeassist.code_assist(project, *self.plugin.get_source_and_offset(),
                    resource=self.plugin.get_rope_resource(project), maxfixes=3))
                    
        except Exception, e:
            import traceback
            traceback.print_exc()
            self.plugin.editor.message(str(e), 5000) 
            return

        if proposals:
            context.add_proposals(self, [Proposal(p) for p in proposals], True)
        else:
            context.add_proposals(self, [], True)
            self.plugin.editor.message("Can't assist")
        
gobject.type_register(RopeCompletionProvider)
gobject.type_register(Proposal)



import os.path
import os

import gtk

def show_create_file(editor):
    widget = create_widget(editor)
    
    editor.widget.pack_start(widget, False)
    widget.entry.grab_focus()
    widget.entry.set_position(-1)
    widget.show_all()

    editor.push_escape(hide, editor, widget)

def create_widget(editor):
    widget = gtk.HBox(False, 10)
    
    label = gtk.Label("File name:")
    widget.pack_start(label, False)
    
    entry = gtk.Entry()
    entry.set_width_chars(50)
    widget.pack_start(entry, False)
    entry.connect('activate', on_entry_activate, editor, widget)
    entry.connect('key-press-event', on_key_press)
    
    completion = gtk.EntryCompletion()
    model = gtk.ListStore(str)
    completion.set_model(model)
    completion.set_inline_completion(True)
    
    cell = gtk.CellRendererText()
    completion.pack_start(cell)
    completion.add_attribute(cell, 'text', 0)
    
    completion.set_match_func(match_func)

    entry.set_completion(completion)
    
    path = os.path.dirname(editor.uri)
    entry.set_text(path + '/')
    fill_model(model, path)

    entry.connect('changed', on_entry_changed, model)
    
    widget.entry = entry
    
    return widget

def hide(editor, widget):
    if widget and widget.get_parent():
        editor.widget.remove(widget)
        widget.destroy()
    editor.view.grab_focus()

def on_entry_activate(entry, editor, widget):
    filename = entry.get_text()
    if os.path.exists(filename):
        editor.message("%s already exists" % filename)
    else:
        hide(editor, widget)
        editor.open_file(filename)
    
def on_entry_changed(entry, model):
    path = os.path.dirname(entry.get_text())
    if path != model.last_path:
        fill_model(model, path, entry)
    
def on_key_press(sender, event):
    if event.keyval in (gtk.keysyms.ISO_Left_Tab, gtk.keysyms.Up, gtk.keysyms.Down):
        return True
    
    if event.keyval == gtk.keysyms.Tab:
        matches = get_matches(sender)
        if len(matches) == 1:
            sender.set_text(matches[0])
            sender.set_position(-1)
        elif len(matches) > 1:
            idx = -1
            text = sender.get_text()
            for i, m in enumerate(matches):
                if m == text:
                    idx = i
                    break
                    
            old_pos = get_pos(sender)
            sender.handler_block_by_func(on_entry_changed)
            sender.set_text(matches[(idx + 1) % len(matches)])
            sender.set_position(old_pos)
            sender.select_region(old_pos, -1)
            sender.handler_unblock_by_func(on_entry_changed)
             
        return True        
        
    return False

def get_pos(entry):
    try:
        start, end = entry.get_selection_bounds()
    except ValueError:
        start = entry.get_position()
    
    return start
    
def get_key(entry):
    return os.path.basename(entry.get_text()[:get_pos(entry)])

def match_func(completion, key, iter):
    key = get_key(completion.get_entry())
    model = completion.get_model()
    text = model.get_value(iter, 0)
    if text and text.startswith(key):
        return True
    
    return False

def get_matches(entry):
    model = entry.get_completion().get_model()
    key = get_key(entry)
    return sorted([os.path.join(model.last_path, text) + '/'
        for (text,) in model if text.startswith(key)])

def fill_model(model, path, entry=None):
    model.clear()
    model.last_path = path
    try:
        for p in os.listdir(path):
            fullpath = os.path.join(path, p)
            if os.path.isdir(fullpath):
                model.append((p,))
    except OSError:
        pass


author = 'Anton Bobrov<bobrov@vl.ru>'
name = 'Search'
desc = 'Searches words in document'

import gtk
import gtksourceview2

from snaked.util import idle

active_widgets = {}

def init(manager):
    manager.add_shortcut('search', '<ctrl>f', 'Edit', 'Search', search)
    manager.add_shortcut('find-next', '<ctrl>j', 'Edit', 'Find next', find_next)
    manager.add_shortcut('find-prev', '<ctrl>k', 'Edit', 'Find prev', find_prev)

def search(editor):
    if editor in active_widgets:
        widget = active_widgets[editor]
        widget.entry.grab_focus()
    else:
        widget = create_widget(editor)
        active_widgets[editor] = widget
        editor.widget.pack_start(widget, False)
        widget.entry.grab_focus()
        widget.show_all()

def do_find(editor, search_func, dir, start_from=None):
    if editor not in active_widgets:
        return
    
    search = active_widgets[editor].entry.get_text()
    
    iter = start_from
    if not iter:
        if editor.buffer.get_has_selection() and dir == 1:
            iter = editor.buffer.get_iter_at_mark(editor.buffer.get_selection_bound())
        else:
            iter = editor.cursor
            
    bounds = search_func(iter, search,
        gtksourceview2.SEARCH_VISIBLE_ONLY | gtksourceview2.SEARCH_CASE_INSENSITIVE)

    if bounds:
        editor.buffer.select_range(bounds[1], bounds[0])
        editor.view.scroll_to_iter(bounds[0], 0.001, use_align=True, xalign=1.0)
        if start_from:
            editor.message('Wrap search', 800)
        
        return True
    elif not start_from:
        return do_find(editor, search_func, dir, editor.buffer.get_bounds()[dir])
    else:
        editor.message('Text not found')
    
    return False
    
def find_prev(editor):
    do_find(editor, gtksourceview2.iter_backward_search, 1) 

def find_next(editor, grab_focus=False):
    if do_find(editor, gtksourceview2.iter_forward_search, 0) and grab_focus:
        editor.view.grab_focus() 

def create_widget(editor):
    widget = gtk.HBox(False, 10)
    
    label = gtk.Label()
    label.set_text('Search:')
    widget.pack_start(label, False)
    
    entry = gtk.Entry()
    widget.pack_start(entry, False)
    entry.connect('activate', on_search_activate, editor, widget)
    entry.connect('key-press-event', on_key_press, editor, widget)
    
    widget.entry = entry
    
    return widget

def get_tag(editor):
    table = editor.buffer.get_tag_table()
    tag = table.lookup('search')
    if not tag:
        style = editor.buffer.get_style_scheme().get_style('search-match')
        tag = editor.buffer.create_tag('search')
        
        if style.props.background_set:
            tag.props.background = style.props.background
            
        if style.props.foreground_set:
            tag.props.foreground = style.props.foreground
            
    return tag

def delete_all_marks(editor):
    start, end = editor.buffer.get_bounds()
    if editor.buffer.get_tag_table().lookup('search'):
        editor.buffer.remove_tag_by_name('search', start, end)

def mark_occurences(editor, search):
    cursor = editor.buffer.get_bounds()[0]
    
    while True:
        bounds = gtksourceview2.iter_forward_search(cursor, search,
            gtksourceview2.SEARCH_VISIBLE_ONLY | gtksourceview2.SEARCH_CASE_INSENSITIVE)
        
        if not bounds:
            return
        
        editor.buffer.apply_tag(get_tag(editor), *bounds)
        
        cursor = bounds[1]

def on_search_activate(sender, editor, widget):
    delete_all_marks(editor)
    idle(mark_occurences, editor, sender.get_text())
    editor.push_escape(hide, editor, widget)
    find_next(editor, True)

def on_key_press(sender, event, editor, widget):
    if event.keyval == gtk.keysyms.Escape:
        idle(hide, editor, widget)
        return True
    
    return False
    
def hide(editor, widget):
    delete_all_marks(editor)
    
    try:
        del active_widgets[editor]
    except KeyError:
        pass
        
    if widget and widget.get_parent():
        editor.widget.remove(widget)
        widget.destroy()

    editor.view.grab_focus()



import os

import gtk
import gio

from snaked.signals import connect_external, connect_all
from snaked.util import idle, refresh_gui, lazy_property

from .ropehints import FileHintDb

class Plugin(object):
    def __init__(self, editor):
        self.editor = editor
        idle(connect_all, self, view=editor.view)
        idle(self.init_completion)

    def init_completion(self):
        provider = self.completion_provider
        self.editor.view.get_completion().add_provider(provider)
        
    def close(self):
        if hasattr(self, '__project'):
            self.project.close()
            self.hints_monitor.cancel()
    
    def refresh_hints(self, filemonitor, file, other_file, event, db):
        if event in (gio.FILE_MONITOR_EVENT_CHANGES_DONE_HINT, gio.FILE_MONITOR_EVENT_CREATED):        
            db.refresh()
    
    @lazy_property
    def project(self):
        if not os.access(self.editor.uri, os.W_OK):
            root = '/tmp'
        else:
            root = self.editor.project_root
        
        if root:
            from rope.base.project import Project
            project = Project(root)
            db = FileHintDb(project)

            self.hints_monitor = gio.File(db.hints_filename).monitor_file()
            self.hints_monitor.connect('changed', self.refresh_hints, db)
            
            return project
        else:
            return None           
    
    @lazy_property
    def completion_provider(self):
        import complete
        return complete.RopeCompletionProvider(self)

    def get_rope_resource(self, project, uri=None):
        from rope.base import libutils    
        uri = uri or self.editor.uri
        return libutils.path_to_resource(project, uri)

    def get_source_and_offset(self):
        offset = self.editor.cursor.get_offset()
        source = self.editor.text
        
        if not isinstance(source, unicode):
            source = source.decode('utf8')
        
        return source, offset

    def goto_line(self, editor, line):
        refresh_gui()
        edit = editor.buffer
        iterator = edit.get_iter_at_line(line - 1)
        edit.place_cursor(iterator)
        editor.view.scroll_to_iter(iterator, 0.001, use_align=True, xalign=1.0)

    def goto_definition(self):
        project = self.project
        if not project:
            project = getattr(self.editor, 'ropeproject', None)
            if not project:
                self.editor.message("Can't find project path")
                return
        
        project.validate()
                 
        current_resource = self.get_rope_resource(project) 
        
        from rope.contrib import codeassist

        try:
            resource, line = codeassist.get_definition_location(
                project, *self.get_source_and_offset(),
                resource=current_resource, maxfixes=3)
        except Exception, e:
            import traceback
            traceback.print_exc()
            self.editor.message(str(e), 5000)
            return
        
        if resource and resource.real_path == current_resource.real_path:
            resource = None
            
        if resource:
            uri = resource.real_path
            editor = self.editor.open_file(uri)
            editor.ropeproject = project 
            self.goto_line(editor, line)
        else:
            if line:
                self.goto_line(self.editor, line)
            else:
                self.editor.message("Unknown definition")


    @connect_external('view', 'key-press-event')
    def on_textview_key_press_event(self, sender, event):
        if event.keyval != gtk.keysyms.Return:
            return False
            
        cursor = self.editor.cursor
        line_start = cursor.copy()
        line_start.set_line(line_start.get_line())
        
        text = line_start.get_text(cursor).strip()
        if text and text[-1] == ':':
            end = line_start.copy()
            end.forward_word_end()
            end.backward_word_start()
            ws = line_start.get_text(end)

            if self.editor.view.get_insert_spaces_instead_of_tabs():
                tab = u' ' * self.editor.view.get_tab_width()
            else:
                tab = u'\t'
                
            self.editor.buffer.begin_user_action()
            self.editor.buffer.insert(cursor, u'\n' + ws + tab)
            self.editor.buffer.end_user_action()
            
            if cursor.is_end():
                idle(self.editor.view.scroll_to_iter, cursor, 0.001, use_align=True, xalign=1.0)
            
            return True
        
        return False

    @connect_external('view', 'backspace')
    def on_textview_backspace(self, *args):
        cursor = self.editor.cursor
        
        if cursor.starts_line():
            return False
        
        start = cursor.copy()
        start.set_line(start.get_line())
            
        text = start.get_text(cursor)
        
        if text.strip():
            return False
            
        delete_from = cursor.copy()
        if text[-1] == u'\t': 
            delete_from.backward_char()
        else:
            delete_from.backward_chars(self.editor.view.get_tab_width() - 1)
        
        if delete_from.get_line() != start.get_line():
            delete_from = start

        if delete_from.equal(start):
            delete_from.forward_char()

        self.editor.buffer.begin_user_action()
        self.editor.buffer.delete(delete_from, cursor)
        self.editor.buffer.end_user_action()

        return True



author = 'Anton Bobrov<bobrov@vl.ru>'
name = 'Quick open'
desc = 'Allows one to open files in project very fast and easy'

dialog = None
        
def init(manager):
    manager.add_shortcut('quick-open', '<ctrl><alt>r', 'File', "Shows quick open dialog", activate)

def editor_opened(editor):
    import settings
    root = editor.project_root
    if root and root not in settings.recent_projects:            
        settings.recent_projects.append(root)                
    
def activate(editor):
    global dialog
    if not dialog:
        from gui import QuickOpenDialog
        dialog = QuickOpenDialog()

    dialog.show(editor)
        
def quit():
    global gui
    gui = None



author = 'Anton Bobrov<bobrov@vl.ru>'
name = 'Edit and Select'
desc = 'Various edit shortcuts'

import gtk

def init(manager):
    manager.add_shortcut('delete-line', '<ctrl>d', 'Edit', 'Deletes current line', delete_line)
    manager.add_shortcut('smart-select', '<alt>w', 'Selection', 'Smart anything selection', smart_select)
    manager.add_shortcut('show_offset', '<ctrl><alt>o', 'Edit', 'Show cursor offset', show_offset)

def delete_line(editor):
    from util import get_line_bounds
    bounds = get_line_bounds(editor.cursor)
    clipboard = editor.view.get_clipboard(gtk.gdk.SELECTION_CLIPBOARD)
    editor.buffer.select_range(*bounds)
    editor.buffer.copy_clipboard(clipboard)
    
    editor.buffer.begin_user_action()
    editor.buffer.delete(*bounds)
    editor.buffer.end_user_action()

def smart_select(editor):
    from util import select_range
    from smart_select import get_smart_select
    select_range(editor.buffer, *get_smart_select(editor))
    
def show_offset(editor):
    editor.message(str(editor.cursor.get_offset()), 3000)



import gtk
import os.path

from snaked.core.prefs import get_settings_path

registered_shortcuts = {}
names_by_key = {}
default_shortcuts = {}

def get_path_by_name(name):
    return "<Snaked-Editor>/%s/%s" % (registered_shortcuts[name][0], name)

def get_registered_shortcuts():
    result = []
    
    def func(path, key, mod, changed):
        result.append((path, key, mod))
        
    gtk.accel_map_foreach_unfiltered(func)
    
    return result
    
def refresh_names_by_key():
    for path, key, mod in get_registered_shortcuts():
        names_by_key[(key, mod)] = path

def get_path_by_key(key, mod):
    try:
        return names_by_key[(key, mod)]
    except KeyError:
        refresh_names_by_key()
        return names_by_key[(key, mod)]

def register_shortcut(name, accel, category, desc):
    registered_shortcuts[name] = category, desc
    key, modifier = gtk.accelerator_parse(accel)
    path = get_path_by_name(name)
    default_shortcuts[path] = (key, modifier)
    gtk.accel_map_add_entry(path, key, modifier)    
    
    return path

def save_shortcuts():
    gtk.accel_map_save(get_settings_path('keys.conf'))
    
def load_shortcuts():
    config = get_settings_path('keys.conf')
    if os.path.exists(config):
        gtk.accel_map_load(config)

class Shortcut(object):
    def __init__(self, name, accel, category, desc):
        self.name = name
        self.accel = accel
        self.category = category
        self.desc = desc

    @property
    def keymod(self):
        try:
            return self.__keymod
        except AttributeError:
            self.__keymod = gtk.accelerator_parse(self.accel)
            return self.__keymod


class ShortcutActivator(object):
    def __init__(self, window):
        self.window = window
        self.accel_group = gtk.AccelGroup()
        self.window.add_accel_group(self.accel_group)
        
        self.shortcuts = {}
        self.pathes = {}
        
    def bind(self, accel, callback, *args):
        key, modifier = gtk.accelerator_parse(accel)
        self.shortcuts[(key, modifier)] = callback, args
        
        self.accel_group.connect_group(key, modifier, gtk.ACCEL_VISIBLE, self.activate)
    
    def bind_to_name(self, name, callback, *args):
        path = get_path_by_name(name)
        self.pathes[path] = callback, args
        self.accel_group.connect_by_path(path, self.activate)
    
    def get_callback_and_args(self, *key):
        try:
            return self.shortcuts[key]
        except KeyError:
            return self.pathes[get_path_by_key(*key)]
    
    def activate(self, group, window, key, modifier):
        cb, args = self.get_callback_and_args(key, modifier)
        result = cb(*args)
        return result is None or result


class ContextShortcutActivator(ShortcutActivator):
    def __init__(self, window, context):
        super(ContextShortcutActivator, self).__init__(window)
        self.context = context

    def activate(self, group, window, key, modifier):
        ctx = self.context()
        cb, args = self.get_callback_and_args(key, modifier)
        
        if hasattr(cb, 'provide_key'):
            return cb(key, modifier, *(ctx + args))
        else:
            result = cb(*(ctx + args))
            return result is None or result



author = 'Anton Bobrov<bobrov@vl.ru>'
name = 'Python flakes'
desc = 'Basic python linter'

import gtk
import pango
import os.path

langs = ['python']

def editor_opened(editor):
    editor.connect('file-saved', on_file_saved)
    if editor.uri and os.path.exists(editor.uri):    
        add_job(editor)
    
def on_file_saved(editor):
    add_job(editor)
    
def get_tag(editor):
    table = editor.buffer.get_tag_table()
    tag = table.lookup('flakes')
    if not tag:
        tag = editor.buffer.create_tag('flakes', underline=pango.UNDERLINE_ERROR)
    
    return tag

def mark_problems(editor, problems):
    start, end = editor.buffer.get_bounds()
    
    if editor.buffer.get_tag_table().lookup('flakes'):
        editor.buffer.remove_tag_by_name('flakes', start, end)
    
    for line, name, message in problems:
        iter = editor.buffer.get_iter_at_line(line-1)
        nstart, nend = iter.forward_search(name, gtk.TEXT_SEARCH_VISIBLE_ONLY)
        editor.buffer.apply_tag(get_tag(editor), nstart, nend) 
        
def get_problem_list(filename):
    import subprocess
    import re
    
    stdout, stderr = subprocess.Popen(['/usr/bin/env', 'pyflakes', filename],
        stdout=subprocess.PIPE, stderr=subprocess.PIPE).communicate()

    if stderr:
        raise Exception('python_flakes plugin: ' + stderr)
    
    result = []
    match_name = re.compile(r"'(.*?)'")
    for m in re.finditer(r"(?m)^.*?:(?P<line>\d+):\s*(?P<message>.*$)", stdout):
        line, message = m.group('line', 'message')
        name = match_name.search(message)
        if not name:
            raise Exception("Can't parse variable name in " + message)
        
        result.append((int(line), name.group(1), message))
        
    return result

def add_job(editor):
    from threading import Thread
    from snaked.util import idle
    
    def job():
        try:
            problems = get_problem_list(editor.uri)
        except Exception, e:
            idle(editor.message, str(e), 5000)
            return

        idle(mark_problems, editor, problems)
        
    Thread(target=job).start()



author = 'Anton Bobrov<bobrov@vl.ru>'
name = 'Hash comment'
desc = '(Un)Comments line or selection with hashes'

import gtk

langs = ['python', 'sh', 'ruby', 'perl']

def init(manager):
    manager.add_shortcut('comment-code', '<ctrl>slash', 'Edit',
        'Comments/uncomments current line or selection', comment)

#Action
def comment(editor):
    r = get_bounds(editor)
    traversor = make_line_traversor(editor.buffer, r)
            
    if all(line_is_hashed(l) for l in traversor()):
        uncomment_range(editor, traversor)
    else:
        comment_range(editor, traversor)

def make_line_traversor(buffer, r):
    start, end = r
    start, stop = start.get_line(), end.get_line() + 1
    def inner():
        for i in xrange(start, stop):
            yield buffer.get_iter_at_line(i)
    
    return inner

def get_bounds(editor):
    if editor.buffer.get_has_selection():
        start, end = editor.buffer.get_selection_bounds()        
        if start.ends_line():
            start.set_line(start.get_line() + 1)
            
        if end.starts_line():
            end.set_line(end.get_line() - 1)
        
        return start, end
    else:
        cursor = editor.cursor
        return cursor, cursor.copy()

def line_is_hashed(iter):
    text = get_line_text(iter).strip()
    return not text or text[0] == u'#'

def get_line_text(iter):    
    if not iter.starts_line():
        iter = iter.copy()
        iter.set_line(iter.get_line())

    end = iter.copy()
    if not end.ends_line():
        end.forward_to_line_end()

    return iter.get_text(end)
    
def line_is_empty(iter):
    return get_line_text(iter).strip() == u''
    
def comment_range(editor, traversor):
    editor.buffer.begin_user_action()
    for iter in traversor():
        editor.buffer.insert(iter, u'#')

    editor.buffer.end_user_action()

def uncomment_range(editor, traversor):
    editor.buffer.begin_user_action()
    for iter in traversor():
        if not line_is_empty(iter):
            editor.buffer.delete(*iter.forward_search(u'#', gtk.TEXT_SEARCH_VISIBLE_ONLY))
                   
    editor.buffer.end_user_action()



import os
import shutil
from os.path import join, dirname, realpath, abspath, exists

import gobject
import gtk

from snaked.signals.signals import Handler
from snaked.signals import weak_connect

def idle_callback(callable, args):
    args, kwargs = args
    callable(*args, **kwargs)
    return False

def idle(callable, *args, **kwargs):
    return gobject.idle_add(idle_callback, callable, (args, kwargs))

def save_file(filename, data, encoding):
    tmpfilename = realpath(filename) + '.bak'
    
    try:
        f = open(tmpfilename, 'w')
    except IOError:
        dname = dirname(tmpfilename)
        if not exists(dname):
            os.makedirs(dname, mode=0755)
            f = open(tmpfilename, 'w')
                    
    f.write(data.encode(encoding))
    f.close()
    
    if exists(filename):
        shutil.copymode(filename, tmpfilename)
    
    os.rename(tmpfilename, filename)

def connect(sender, signal, obj, attr, idle=False, after=False):
    return Handler(weak_connect(
        sender, signal, obj, attr, idle=idle, after=after), sender)

def join_to_file_dir(filename, *args):
    return join(dirname(filename), *args)

def get_project_root(filename):
    path = dirname(abspath(filename))
    magic_pathes = ['.git', '.hg', '.bzr', '.ropeproject', '.snaked_project']
    
    while True:
        if any(exists(join(path, p)) for p in magic_pathes):
            return path
    
        parent = dirname(path)
        if parent == path:
            return None
    
        path = parent
    
def open_mime(filename):
    import subprocess
    subprocess.Popen(['/usr/bin/env', 'xdg-open', filename]).poll()


def refresh_gui():
    while gtk.events_pending():
        gtk.main_iteration_do(block=False)

def single_ref(func):
    real_name = '__' + func.__name__
    holder_name = func.__name__ + '_holder'
    
    def inner(self):
        try:
            return getattr(self, real_name)
        except AttributeError:
            pass
            
        cls = self.__class__
        if not hasattr(cls, holder_name) or not getattr(cls, holder_name)():
            import weakref
            var = func(self)
            setattr(cls, holder_name, weakref.ref(var))
        
        setattr(self, real_name, getattr(cls, holder_name)())
        return getattr(self, real_name)
        
    return property(inner)

def lazy_property(func):
    real_name = '__' + func.__name__
    
    def inner(self):
        try:
            return getattr(self, real_name)
        except AttributeError:
            pass
        
        var = func(self)
        setattr(self, real_name, var)
        return var
        
    return property(inner)


class BuilderAware(object):
    def __init__(self, glade_file):
        self.gtk_builder = gtk.Builder()
        self.gtk_builder.add_from_file(glade_file)
        self.gtk_builder.connect_signals(self)
    
    def __getattr__(self, name):
        obj = self.gtk_builder.get_object(name)
        if not obj:
            raise AttributeError('Builder have no %s object' % name)
            
        setattr(self, name, obj)
        return obj



def get_manager():
    from optparse import OptionParser
    import sys
    
    parser = OptionParser()
    parser.add_option('-s', '--session', dest='session', help="Open snaked with specified session")
    parser.add_option('-w', '--windowed', action="store_true", 
        dest='windowed', default = False, help="Open separate editor window instead tab")
    options, args = parser.parse_args()

    import gobject
    gobject.threads_init()
    
    if options.windowed:
        from .windowed import WindowedEditorManager as WM
    else:
        from .tabbed import TabbedEditorManager as WM

    manager = WM()

    opened_files = []
    
    session_files = []
    if options.session:
        session_files = manager.get_session_files(options.session)
    
    for f in session_files + args:
        if f not in opened_files:    
            manager.open(f)
            opened_files.append(f)

    if not manager.editors:
        print >> sys.stderr, 'You must specify at least one file to edit'
        sys.exit(1)
        
    return manager
            
def run():
    manager = get_manager()

    import gtk
    
    try:    
        gtk.main()
    except KeyboardInterrupt:
        manager.quit()



import gtk

from snaked.core.shortcuts import ContextShortcutActivator
import snaked.core.editor

class TabbedEditorManager(snaked.core.editor.EditorManager):
    def __init__(self):
        super(TabbedEditorManager, self).__init__()

        self.window = gtk.Window(gtk.WINDOW_TOPLEVEL)
        self.window.connect('delete-event', self.on_delete_event)

        self.window.set_property('default-width', 800)
        self.window.set_property('default-height', 500)
    
        self.activator = ContextShortcutActivator(self.window, self.get_context)
        
        self.note = gtk.Notebook()
        self.note.set_property('tab-hborder', 10)
        self.note.set_property('homogeneous', False)
        self.note.connect_after('switch-page', self.on_switch_page)
        self.window.add(self.note)
        
        self.window.show_all()
    
    def get_context(self):
        widget = self.note.get_nth_page(self.note.get_current_page())
        for e in self.editors:
            if e.widget is widget:
                return (e,)

        raise Exception('Editor not found')

    def manage_editor(self, editor):
        label = gtk.Label('Unknown')
        self.note.append_page(editor.widget, label)
        self.focus_editor(editor)
        editor.view.grab_focus()
       
    def focus_editor(self, editor):
        idx = self.note.page_num(editor.widget)
        self.note.set_current_page(idx)

    def update_top_level_title(self):
        idx = self.note.get_current_page()
        if idx < 0:
            return
        
        title = self.note.get_tab_label_text(self.note.get_nth_page(idx))
        if title is not None:
            self.window.set_title(title)        
                
    def set_editor_title(self, editor, title):
        self.note.set_tab_label_text(editor.widget, title)
        if self.note.get_current_page() == self.note.page_num(editor.widget):
            self.update_top_level_title()

    def on_delete_event(self, *args):
        self.quit()

    def close_editor(self, editor):
        idx = self.note.page_num(editor.widget)
        self.note.remove_page(idx)
        editor.editor_closed.emit()

    def set_editor_shortcuts(self, editor):
        self.plugin_manager.bind_shortcuts(self.activator, editor)

        if hasattr(self, 'editor_shortcuts_binded'):
            return
        
        self.editor_shortcuts_binded = True

        self.activator.bind_to_name('quit', self.quit)
        self.activator.bind_to_name('close-window', self.close_editor)
        self.activator.bind_to_name('save', self.save)
        self.activator.bind_to_name('next-editor', self.switch_to, 1)
        self.activator.bind_to_name('prev-editor', self.switch_to, -1)
        self.activator.bind_to_name('new-file', self.new_file_action)

        self.activator.bind_to_name('show-preferences', self.show_preferences)

        self.activator.bind('Escape', self.process_escape)

    def quit(self, *args):
        self.window.hide()
        super(TabbedEditorManager, self).quit()

    def save(self, editor):
        editor.save()

    def set_transient_for(self, editor, window):
        window.set_transient_for(self.window)

    def on_switch_page(self, *args):
        self.update_top_level_title()
        
    def switch_to(self, editor, dir):
        idx = ( self.note.get_current_page() + dir ) % self.note.get_n_pages()
        self.note.set_current_page(idx)



author = 'Anton Bobrov<bobrov@vl.ru>'
name = 'Goto dir'
desc = "Opens file's directory"

def init(manager):
    manager.add_shortcut('goto-dir', '<ctrl><alt>l', 'File', "Opens file's directory", goto_dir)
    
def goto_dir(editor):
    import subprocess
    import os.path
    
    subprocess.Popen(['/usr/bin/env', 'xdg-open', os.path.dirname(editor.uri)]).poll()
    editor.message('File manager started', 1000)


import os.path
import re

import weakref

import rope.base.oi.soi
import rope.base.pyobjects
import rope.base.pynames
from rope.base import exceptions
from rope.base.pyobjectsdef import PyModule, PyPackage

def update_type_for(project, source, resource):
    pass
    
def infer_parameter_objects_with_hints(func):
    def inner(pyfunction):
        params_types = func(pyfunction)
        
        try:
            hintdb = pyfunction.pycore.hintdb
        except AttributeError:
            return params_types
        
        param_names = pyfunction.get_param_names(False)
        for i, name in enumerate(param_names):
            ptype = hintdb.get_function_param_type(pyfunction, name)
            if ptype != None:
                params_types[i] = ptype
        
        return params_types
        
    return inner

rope.base.oi.soi.infer_parameter_objects = infer_parameter_objects_with_hints(
    rope.base.oi.soi.infer_parameter_objects)


def infer_returned_object_with_hints(func):
    def inner(pyfunction, args):
        try:
            hintdb = pyfunction.pycore.hintdb
        except AttributeError:
            return func(pyfunction, args)
        
        rtype = hintdb.get_function_param_type(pyfunction, 'return')
        if rtype is None:
            rtype = func(pyfunction, args)
        
        return rtype
        
    return inner

rope.base.oi.soi.infer_returned_object = infer_returned_object_with_hints(
    rope.base.oi.soi.infer_returned_object)


def get_module_attribute_with_hints(func, what):
    def inner(self, name):
        #print what, self.pycore.modname(self.resource), name

        try:
            hintdb = self.pycore.hintdb
        except AttributeError:
            return func(self, name)

        try:
            original_pyname = func(self, name)
        except exceptions.AttributeNotFoundError:
            original_pyname = None
        
        result = hintdb.get_module_attribute(self, name, original_pyname)
        if not result:
            raise exceptions.AttributeNotFoundError()
        else:
            return result
        
    return inner
        
PyModule.get_attribute = get_module_attribute_with_hints(PyModule.get_attribute, 'mod')
PyPackage.get_attribute = get_module_attribute_with_hints(PyPackage.get_attribute, 'pkg')

class HintDb(object):
    def __init__(self, project):
        self.type_cache = {}
        self.module_attrs_cache = weakref.WeakKeyDictionary()
        project.pycore.hintdb = self
        
    def get_function_param_type(self, pyfunc, name):
        scope_path = self.get_scope_path(pyfunc.get_scope())
        type_name = self.find_type_for(scope_path, name)
        if type_name:
            return self.get_type(pyfunc.pycore, type_name).get_object()
        else:
            return None
        
    def get_scope_path(self, scope):
        result = []
        current_scope = scope
        while current_scope is not None:
            pyobj = current_scope.pyobject
            if isinstance(pyobj, PyModule):
                name = pyobj.pycore.modname(pyobj.resource)
            else:
                name = pyobj.get_name()
        
            result.insert(0, name)
            current_scope = current_scope.parent
        
        return '.'.join(result)
        
    def find_type_for(scope_path, name):
        return None
        
    def get_type(self, pycore, type_name):
        try:
            return self.type_cache[type_name]
        except KeyError:
            pass
        
        module, sep, name = type_name.strip('()').rpartition('.')
        if module:
            module = pycore.get_module(module)
            obj = module[name]
        else:
            obj = pycore.get_module(name)
        
        self.type_cache[type_name] = obj
        return obj
        
    def get_module_attribute(self, pymodule, name, original_pyname):
        try:
            return self.module_attrs_cache[pymodule][name]
        except KeyError:
            pass

        scope_path = pymodule.pycore.modname(pymodule.resource)
        type_name = self.find_type_for(scope_path, name)
        if type_name:
            type = self.get_type(pymodule.pycore, type_name)
        else:
            type = None
        
        if type:
            if type_name.endswith('()'):
                type = type.get_object()
                
                obj = rope.base.pyobjects.PyObject(type)
                if isinstance(original_pyname, rope.base.pynamesdef.AssignedName):
                    pyname = rope.base.pynamesdef.AssignedName(original_pyname.lineno,
                        original_pyname.module, obj)
                else:
                    pyname = rope.base.pynames.DefinedName(obj)
            else:
                pyname = type
        else:
            pyname = original_pyname
        
        self.module_attrs_cache.setdefault(pymodule, {})[name] = pyname
        return pyname

        
class ReHintDb(HintDb):
    def __init__(self, project):
        super(ReHintDb, self).__init__(project)
        self.db = []
        
    def add_hint(self, scope, name, object_type):
        self.db.append((re.compile(scope), re.compile(name), object_type))  

    def find_type_for(self, scope_path, name):
        for scope, vname, otype in self.db:
            if scope.search(scope_path) and vname.search(name):
                return otype
                
        return None
        
class FileHintDb(ReHintDb):
    def __init__(self, project):
        super(FileHintDb, self).__init__(project)
        self.hints_filename = os.path.join(project.ropefolder.real_path, 'hints')
        self.load_hints()
        
    def refresh(self):
        self.load_hints()
        
    def load_hints(self):
        self.db[:] = []
        self.module_attrs_cache.clear()
        
        if os.path.exists(self.hints_filename):
            with open(self.hints_filename) as f:
                for l in f:
                    try:
                        scope, name, type = l.strip().split()
                        self.add_hint(scope, name, type)
                    except ValueError:
                        continue